structure Test = 
struct


(** Required code to tie the lexer and the parser together **)
structure CalcLrVals = CalcLrValsFun (structure Token = LrParser.Token)

structure CalcLex = CalcLexFun (structure Tokens = CalcLrVals.Tokens)

structure CalcParser = 
  Join (structure LrParser = LrParser
        structure ParserData = CalcLrVals.ParserData
        structure Lex = CalcLex)

fun invoke lexstream =
    let fun print_error (s,i:int,_) =
            TextIO.output(TextIO.stdOut,
                          "Error, line " ^ (Int.toString i) ^ ", " ^ s ^ "\n")
     in CalcParser.parse(0,lexstream,print_error,())
    end

exception Internal_error
exception Failure

(* flag_catch_errors: exceptions are caught and reported if true, left
   uncaught if false *)
val flag_catch_errors = true

(* evaluate a term in deBruijn form *)
fun eval t =
    let 
        (* do the evaluation *)
        val t = if flag_catch_errors then
                    SOME(Ast.eval t) handle 
                    Ast.Step_error => (print ("Error: type error\n"); NONE)
                else 
                    SOME(Ast.eval t) 
        
        (* print the results, if we succeeded *)            
        val t = case t of 
                    SOME t => ((print ("evals-to  : " ^ (Ast.pretty_print_db t) ^ "\n")); t)
                  | NONE => raise Failure
    in
        (Ast.pretty_print_db t)
    end    

(* parse, convert and evaluate a term *)
fun run t =
    let
        (* parse and ensure it gave us a well-formed ast (no indexes, etc) *)
        val _ = print ("read      : " ^ (Ast.pretty_print t) ^ "\n")
        
        (* convert and ensure it worked *)
        val t = if flag_catch_errors then 
                    SOME(Ast.to_debruijn t) handle 
                    Ast.Free_variable_error => (print ("Error: free-variable error\n"); NONE)
                else
                    SOME(Ast.to_debruijn t)
    in 
        case t of
            SOME t => let
                val _ = print ("converted : " ^ (Ast.pretty_print_db t) ^ "\n")
            in eval t end
          | NONE => raise Failure
    end


(* parse using a lexer *)
fun parse_from_lexer lexer =
    let val dummyEOF = CalcLrVals.Tokens.EOF(0,0)
        val dummySTOP = CalcLrVals.Tokens.STOP(0,0)
        fun loop lexer =
            let val (result,lexer) = invoke lexer
                val (nextToken,lexer) = CalcParser.Stream.get lexer
                val result = case result
                              of SOME r => run r	
                               | NONE => raise Failure
            in 
                if CalcParser.sameToken(nextToken,dummyEOF) 
		   orelse
		   CalcParser.sameToken(nextToken,dummySTOP) 
	        then result
               else loop lexer
            end
    in loop lexer
    end    

(* parse a string *)
fun parse_str str =
    let val lexer = CalcParser.makeLexer (fn _ => str) 
    in 
        parse_from_lexer lexer
    end

(* parse from stdin *)
fun parse_from_input () =
    let val lexer = CalcParser.makeLexer
                        (fn _ => valOf (TextIO.inputLine TextIO.stdIn))
        val _ = print ">>>"
    in
        parse_from_lexer lexer
    end

(* by default, read from stdin *)
val parse = parse_from_input


(* TODO, make this the factorial function in our toy language *)
val factorial_function = "lam x . x"


exception Conversion_failed

(* use factorial function (a string, above) to compute factorial of
   given argument, raises Conversion_failed if something goes wrong. *)
fun fact (n : int) =
    let
        val evalstr = "(" ^ factorial_function ^ ")" ^ (Int.toString n) ^ ";;"
        val evalres = parse_str evalstr 
        val fact = case Int.fromString evalres of
                       SOME n => n | NONE => raise Conversion_failed
    in
        fact
    end
end
    
